import java.util.concurrent.atomic.AtomicInteger;

/**
 * Test: Setze einen Zhler auf 0 und starte 10 Threads, die den Zhler 100'000
 * Mal um eins erhhen. Schreib das Resultat in den System.out und gib die Zeit
 * zurck.
 * Drei Zhlvarianten:
 * 1. Ungeschtzt, ein int wird mit ++ hochgezhlt. Die Threads berschreiben
 * sich gegenseitig und es ergibt (oft) ein falsches Resultat.
 * 2. Die Threads werden mit synchronized(lock) synchronisiert. Dies kostet viel
 * Zeit, aber das Resultat ist korrekt.
 * 3. Die Zhlvariable ist ein AtomicInteger. Das Resultat ist korrekt und der
 * Zhlvorgang verglichen mit der synchornisierten Methode schnell.
 */
public class AtomicTest {

	private static int n = 100000;
	private static int counter;
	private static AtomicInteger atomicCounter;
	
	public static void main(String[] args) {
		long minUnprotected = Long.MAX_VALUE;
		long minSynchronized = Long.MAX_VALUE;
		long minAtomic = Long.MAX_VALUE;
		int m = 100;
		
		// Fhre jeden Test 100 Mal aus und speichere den minimalsten Zeitaufwand
		for (int i=0;i<m;i++) {
			minUnprotected = Math.min(minUnprotected,testUnprotected());
		}
		
		for (int i=0;i<m;i++) {
			minSynchronized = Math.min(minSynchronized,testSynchronized());
		}
		
		for (int i=0;i<m;i++) {
			minAtomic = Math.min(minAtomic,testAtomic());
		}
		
		System.out.println("Unprotected min:  "+minUnprotected/1000+" mms");
		System.out.println("Synchronized min: "+minSynchronized/1000+" mms");
		System.out.println("Atomic min:       "+minAtomic/1000+" mms");
	}
	
	private static long testAtomic() {
		atomicCounter = new AtomicInteger(0);
		try {
			long t0 = System.nanoTime();
			Thread[] threads = new Thread[10];
			for (int i=0;i<10;i++) {
				threads[i] = new Thread() {
					public void run() {
						for (int i=0;i<n;i++)
							atomicCounter.incrementAndGet();
					}
				};
				threads[i].start();
			}
			for (int i=0;i<10;i++) {
				threads[i].join();
			}
			long dt = System.nanoTime()-t0;
			System.out.println("AtomicTest result = "+atomicCounter.get()+" , time elapsed: "+dt/1000+" mms");
			return dt;
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
	}
	
	private static long testSynchronized() {
		counter = 0;
		final Object lock = new Object();
		try {
			long t0 = System.nanoTime();
			Thread[] threads = new Thread[10];
			for (int i=0;i<10;i++) {
				threads[i] = new Thread() {
					public void run() {
						for (int i=0;i<n;i++)
							synchronized (lock) {
								counter++;								
							}
					}
				};
				threads[i].start();
			}
			for (int i=0;i<10;i++) {
				threads[i].join();
			}
			long dt = System.nanoTime()-t0;
			System.out.println("SynchronizedTest result = "+counter+" , time elapsed: "+dt/1000+" mms");
			return dt;
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
	}
	
	private static long testUnprotected() {
		counter = 0;
		try {
			long t0 = System.nanoTime();
			Thread[] threads = new Thread[10];
			for (int i=0;i<10;i++) {
				threads[i] = new Thread() {
					public void run() {
						for (int i=0;i<n;i++)
							counter++;
					}
				};
				threads[i].start();
			}
			for (int i=0;i<10;i++) {
				threads[i].join();
			}
			long dt = System.nanoTime()-t0;
			System.out.println("Unprotected result = "+counter+" , time elapsed: "+dt/1000+" mms");
			return dt;
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
	}

}
